"""
test_djdt_flamegraph
----------------------------------

Tests for `djdt_flamegraph` module.
"""

import unittest

from djdt_flamegraph import djdt_flamegraph
from djdt_flamegraph import FlamegraphPanel


class TestDjdtFlamegraph(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_000_something(self):
        pass
